import math
import bisect
import itertools

import numpy as np
import scipy.stats
#import overrides

import measurements.util.calculate

import util.logging
import util.multi_dict

logger = util.logging.logger


class MeasurementsDict(util.multi_dict.MultiDict):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)
        self._year_len = 1


    @property
    def year_len(self):
        try:
            return self._year_len
        except AttributeError:
            return 1


    ## str

    def __str__(self):
        name = '{name} with {number} measurements'.format(name=self.__class__.__name__, number=len(self))
        return name


    ## create

    def _return_items_as_type(self, keys, values, return_type=None):
        if return_type == 'measurements' or return_type == 'measurements_unsorted':
            return_type = 'self_type_unsorted'
        if return_type == 'measurements_sorted':
            return_type = 'self_type_sorted'
        return super()._return_items_as_type(keys, values, return_type=return_type)


    ## transform keys

    @staticmethod
    def categorize_index(index, separation_values, discard_year=False):
        index = list(index)

        ## discard year
        if discard_year:
            index[0] = index[0] % self.year_len

        ## remove right bound of last y box
        if index[2] == 90:
            index[2] = 90 - 10**(-6)

        ## iterate over dimensions
        for i in range(len(separation_values)):

            ## get separation value
            try:
                separation_value = separation_values[i]
            except IndexError:
                separation_value = None

            ## categorize dimension i
            if separation_value is not None:
                # check if sequence
                try:
                    is_sequence = len(separation_value) >= 2
                except TypeError:
                    is_sequence = False

                # separate by value
                if not is_sequence:
                    index[i] = (math.floor(index[i] / separation_value) + 0.5) * separation_value
                # separate by sequence
                else:
                    index_right = bisect.bisect_right(separation_value, index[i])
                    value_left = separation_value[index_right - 1]
                    try:
                        value_right = separation_value[index_right]
                    except IndexError:
                        raise ValueError('Index value %d exceeds range of separation values (right sight of separation values is %d).' % (index[i], value_left))
                    index[i] = (value_left + value_right) / 2.0

        index = tuple(index)
        return index


    def categorize_indices(self, separation_values, discard_year=False):
        if discard_year:
            logger.debug('Indices categorized by separation values %s and discard year.' % str(separation_values))
        else:
            logger.debug('Indices categorized by separation values %s.' % str(separation_values))

        transform_function = lambda index: self.categorize_index(index, separation_values, discard_year=discard_year)

        self.transform_keys(transform_function)


    @staticmethod
    def categorize_index_to_lsm(index, lsm, discard_year=False):
        return MeasurementsDict.categorize_index(index, (self.year_len/lsm.t_dim, 360/lsm.x_dim, 180/lsm.y_dim, lsm.z), discard_year=discard_year)


    def categorize_indices_to_lsm(self, lsm, discard_year=False):
        self.categorize_indices((self.year_len/lsm.t_dim, 360/lsm.x_dim, 180/lsm.y_dim, lsm.z), discard_year=discard_year)


    def coordinates_to_map_indices(self, lsm, int_indices=True):
        logger.debug('Transforming in {} coordinates to map indices of {} with int_indices {}'.format(self, lsm, int_indices))
        self._year_len = lsm.t_dim
        self.transform_keys(lambda point: lsm.coordinate_to_map_index(*point, discard_year=False, int_indices=int_indices))

    def map_indices_to_coordinates(self, lsm):
        logger.debug('Transforming in {} map indices of {} to coordinates'.format(self, lsm))
        self._year_len = 1
        self.transform_keys(lambda index: lsm.map_index_to_coordinate(*index))


    def discard_year(self, discard_year=True):
        if discard_year:
            logger.debug('Discarding year.')

            def transform_function(key):
                key_list = list(key)
                key_list[0] = key[0] % self.year_len
                key = tuple(key_list)
                return key

            self.transform_keys(transform_function)


    def discard_time(self):
        logger.debug('Discarding time.')
        self.dicard_key_dim(0)

    def discard_space(self):
        logger.debug('Discarding space.')
        self.dicard_key_dims((1,2,3))


    ## io

    #@overrides.overrides
    def save(self, file):
        only_dict = self.year_len != 1
        super().save(file, only_dict=only_dict)


    ## transform values

    def normalize(self, same_bounds, min_number_of_values=5):
        logger.debug('Normalizing values with same bounds {} and min measurements {}.'.format(same_bounds, min_number_of_values))

        ## save measurements dict
        value_dict = self.value_dict

        ## get means and standard_deviations
        self.categorize_indices(same_bounds, discard_year=True)
        means = self.means(min_number_of_values=min_number_of_values, return_type='self_type_unsorted')
        standard_deviations = self.standard_deviations(min_number_of_values=min_number_of_values, return_type='self_type_unsorted')

        ## prepare new measurements dict
        self.clear()

        ## iterate
        for (key, value_list) in self._iterate_generator_value_dict(value_dict):
            categorized_key = self.categorize_index(key, same_bounds, discard_year=True)
            try:
                mean = means[categorized_key][0]
                standard_deviation = standard_deviations[categorized_key][0]
                match = True
            except KeyError:
                match = False

            if match and standard_deviation > 0:
                new_value_list = []
                for value in value_list:
                    value_normalized = (value - mean) / standard_deviation
                    self.append_value(key, value_normalized)


    def normalize_with_lsm(self, lsm, min_number_of_values=5):
        same_bounds = (self.year_len/lsm.t_dim, 360/lsm.x_dim, 180/lsm.y_dim, lsm.z)
        return self.normalize(same_bounds, min_number_of_values=min_number_of_values)


    ## filter

    def filter_year(self, year, return_type='self'):
        return self.filter_key_range(0, [year, year+self.year_len-10**(-10)], return_type=return_type)


    def filter_coordinates_near_water(self, lsm, max_land_boxes=0, return_type='self'):
        self.filter_with_boolean_function(lambda point, values: lsm.is_coordinate_near_water(point, max_land_boxes=max_land_boxes), return_type=return_type)


    def filter_same_point_with_bounds(self, point, equal_bounds=(0,0,0,0), discard_year=True, only_one_per_year=True):
        ## equal_bounds is float -> copy values for each dim
        try:
            float(equal_bounds)
            equal_bounds = [equal_bounds]*4
        except TypeError:
            pass

        ## search equal_bound for z if z has sequence
        assert len(equal_bounds) == 4
        equal_bounds = list(equal_bounds)
        try:
            float(equal_bounds[3])
        except TypeError:
            ## get z bound for point z value
            z_bound_found = False
            i = 0
            while not z_bound_found:
                if point[3] <= equal_bounds[3][i+1, 0]:
                    z_bound_found = True
                else:
                    i += 1
                    z_bound_found = i == len(equal_bounds[3]) - 1
            equal_bounds[3] = equal_bounds[3][i, 1]
        equal_bounds = np.array(equal_bounds)

        ## prepare point
        point_base = np.array(point, copy=True)
        if discard_year:
            point_base[0] = point_base[0] % self.year_len


        ## filter all measurements with point in equal bounds
        filtered_points = []
        filtered_results = []

        for (t, t_dict) in  self.value_dict.items():
            if (not discard_year and np.abs(point_base[0] - t) <= equal_bounds[0]) or (discard_year and np.abs(point_base[0] - t % self.year_len) <= equal_bounds[0]):
                for (x, x_dict) in t_dict.items():
                    if np.abs(point_base[1] - x) <= equal_bounds[1]:
                        for (y, y_dict) in x_dict.items():
                            if np.abs(point_base[2] - y) <= equal_bounds[2]:
                                for (z, result_list) in y_dict.items():
                                    if np.abs(point_base[3] - z) <= equal_bounds[3]:
                                        point = (t, x, y, z)
                                        for result in result_list:
                                            filtered_points.append(point)
                                            filtered_results.append(result)

        ## filter only one per year
        measurements_filtered = MeasurementsSamePointsDict()

        if only_one_per_year:
            filtered_points = np.array(filtered_points)
            filtered_results = np.array(filtered_results)

            years = np.unique(np.floor_divide(filtered_points[:,0], self.year_len))
            point_scale = np.copy(point_base)
            for year in years:
                point_scale[0] = point_base[0]%self.year_len + year

                min_index = np.linalg.norm(((filtered_points - point_scale) / equal_bounds), ord=2, axis=1).argmin()
                measurements_filtered.append_value(filtered_points[min_index], filtered_results[min_index])
            number_of_filtered_measurements = len(years)
        else:
            measurements_filtered.append_values(filtered_points, filtered_results)
            number_of_filtered_measurements = len(filtered_results)

        logger.debug('{} results for point {} with equal bounds {} filtered.'.format(number_of_filtered_measurements, point, equal_bounds))

        return measurements_filtered


    def filter_same_point_except_year(self, point):
        ## prepare point
        point = list(point)

        ## filter all measurements with same point
        measurements_filtered = MeasurementsSamePointsDict()

        for (t, t_dict) in  self.value_dict.items():
            if point[0] % self.year_len == t % self.year_len:
                point = point.copy()
                point[0] = t
                try:
                    x_dict = t_dict[point[1]]
                    y_dict = x_dict[point[2]]
                    result_list = y_dict[point[3]]
                except KeyError:
                    result_list = None

                if result_list is not None:
                    measurements_filtered.extend_value_list(point, result_list)

        logger.debug('{} results for point {} filtered.'.format(len(measurements_filtered), point))
        return measurements_filtered


    def filter_same_points_with_same_function(self, filter_same_point_function, min_number_of_values=10):
        assert callable(filter_same_point_function)

        measurements_filtered = MeasurementsSamePointsDict()

        if min_number_of_values < float('inf'):

            for (point, results) in self.iterator_keys_and_value_lists():
                point = list(point)
                point[0] = point[0] % self.year_len

                if not point in measurements_filtered:
                    same_point_measurements = filter_same_point_function(point)

                    transform_function = lambda point, result: (point[0], result)
                    same_point_measurements.transform_values(transform_function)
                    same_point_value_list = same_point_measurements.values()

                    if len(same_point_value_list) >= min_number_of_values:
                        logger.debug('{} values with point {} passed filter.'.format(len(same_point_value_list), point))
                        measurements_filtered.extend_value_list(point, same_point_value_list)

        return measurements_filtered


    def filter_same_points_except_year(self, min_number_of_values=10):
        logger.debug('Filtering results with same indicies with min measurements {}.'.format(min_number_of_values))

        filter_same_point_function = lambda point: self.filter_same_point_except_year(point)
        return self.filter_same_points_with_same_function(filter_same_point_function, min_number_of_values=min_number_of_values)


    def filter_same_points_with_bounds(self, equal_bounds=(0,0,0,0), discard_year=True, only_one_per_year=True, min_number_of_values=10):
        logger.debug('Filtering results with same indicies with equal bound {}, discard year {} and min measurements {}.'.format(equal_bounds, discard_year, min_number_of_values))

        filter_same_point_function = lambda point: self.filter_same_point_with_bounds(point, equal_bounds=equal_bounds, discard_year=discard_year, only_one_per_year=only_one_per_year)
        return self.filter_same_points_with_same_function(filter_same_point_function, min_number_of_values=min_number_of_values)


    ## lsm related calulations

    def mean_weighted_by_box_volume(self, lsm):
        logger.debug('Calculate mean of values weighted by box volumes of land-sea-mask {}.'.format(lsm))
        ## copy
        m = self.copy()
        ## convert to lsm map indices
        m.categorize_indices_to_lsm(lsm)
        m.coordinates_to_map_indices(lsm, discard_year=True, int_indices=True)
        ## get indices and values
        items = m.items()
        indices = items[:, :-1]
        values = items[:, -1]
        ## copy volumes
        volumes = lsm.volume_of_boxes_of_map_indices(indices)
        ## calculate mean
        mean_weighted = (values * volumes).sum() / volumes.sum()
        ## return
        logger.debug('Mean of values weighted by box volumes of land-sea-mask {} is {}.'.format(lsm,mean_weighted))
        return mean_weighted


    ## total correlogram and correlation (autocorrelation)

    def _get_first_dim_shifted(self, measurements_dict_list, shift, same_bound, wrap_around_range=None):
        logger.debug('Getting first dim shifted with shift %f and same bound %f.' % (shift, same_bound))

        if self.sorted:
            if wrap_around_range is not None:
                wrap_around_len = wrap_around_range[1] - wrap_around_range[0]

            shift_list = []

            ## iterate over all dicts
            for (measurements_dict, measurements_dict_shifted) in measurements_dict_list:

                keys_view_shifted = measurements_dict_shifted.keys()
                keys_view_shifted_len = len(keys_view_shifted)

                ## iterate over all unshifted
                for (key, value) in measurements_dict.items():

                    ## calculate desired key bounds
                    key_shifted_desired_lower_bound = measurements.util.calculate.wrap_around_index(key + shift - same_bound, wrap_around_range)
                    key_shifted_desired_upper_bound = measurements.util.calculate.wrap_around_index(key + shift + same_bound, wrap_around_range)
                    key_shifted_desired_lower_bound_index = keys_view_shifted.bisect_left(key_shifted_desired_lower_bound)
                    key_shifted_desired_upper_bound_index = keys_view_shifted.bisect_right(key_shifted_desired_upper_bound)


                    ## if desired keys are available
                    if key_shifted_desired_lower_bound_index != key_shifted_desired_upper_bound_index:
                        if key_shifted_desired_upper_bound_index >= keys_view_shifted_len or keys_view_shifted[key_shifted_desired_upper_bound_index] > key_shifted_desired_upper_bound:
                            key_shifted_desired_upper_bound_index -= 1

                        ## calculate desired key range
                        if key_shifted_desired_lower_bound_index <= key_shifted_desired_upper_bound_index:
                            key_shifted_desired_index_range = range(key_shifted_desired_lower_bound_index, key_shifted_desired_upper_bound_index + 1)
                        else:
                            key_shifted_desired_index_range = itertools.chain(range(key_shifted_desired_lower_bound_index, keys_view_shifted_len), range(0, key_shifted_desired_upper_bound_index + 1))

                        ## insert values with shifted values
                        for key_shifted_desired_index in key_shifted_desired_index_range:
                            key_shifted = keys_view_shifted[key_shifted_desired_index]
                            value_shifted = measurements_dict_shifted[key_shifted]
                            shift_list.append((value, value_shifted))

            return shift_list


        else:
            if wrap_around_range is not None:
                wrap_around_len = wrap_around_range[1] - wrap_around_range[0]

            shift_list = []

            ## iterate over all dicts
            for (measurements_dict, measurements_dict_shifted) in measurements_dict_list:

                ## iterate over all unshifted
                for (i, i_dict) in measurements_dict.items():
                    i_shifted_desired = i + shift
                    if wrap_around_range is not None:
                        i_shifted_desired = measurements.util.calculate.wrap_around_index(i_shifted_desired, wrap_around_range)

                    ## iterate over all shifted
                    for (i_shifted, i_dict_shifted) in measurements_dict_shifted.items():
                        i_diff = abs(i_shifted - i_shifted_desired)
                        if wrap_around_range is not None:
                            i_diff = min(i_diff, wrap_around_len - i_diff)

                        ## insert unshifted in shifted in shift_list
                        if i_diff <= same_bound:
                            shift_list.append((i_dict, i_dict_shifted))

            return shift_list


    def _get_results_together_with_shifted(self, factor, direction, same_bounds, x_range, t_range=None):
        logger.debug('Gathering results with direction %s shifted by factor %f with same bound %s.' % (direction, factor, same_bounds))

        measurements_dict_list = [(self.measurements_dict, self.measurements_dict)]
        dim = len(direction)
        wrap_around_range = (t_range, x_range, None, None)

        ## iterate over dim and search matching shifts
        for i in range(dim):
            measurements_dict_list = self._get_first_dim_shifted(measurements_dict_list, factor * direction[i], same_bounds[i], wrap_around_range=wrap_around_range[i])

        logger.debug('Results gathered.')

        return measurements_dict_list


    def _get_array_from_shift_list(self, shift_list):
        ## calculate length
        n = 0
        for (result_list, result_shifted_list) in shift_list:
            n += len(result_list) * len(result_shifted_list)

        ## set values
        array = np.empty((n, 2))
        i = 0
        for (result_list, result_shifted_list) in shift_list:
            for result in result_list:
                for result_shifted in result_shifted_list:
                    array[i, 0] = result
                    array[i, 1] = result_shifted
                    i += 1

        assert i == n

        return array


    def _calculate_total_correlation_from_shift_list(self, shift_list, is_normalized=False):
        if not is_normalized:
            #TODO mean and sd for each result list
            shift_array = self._get_array_from_shift_list(shift_list)
            number = shift_array.shape[0]


            x = shift_array[:,0]
            y = shift_array[:,1]

            mean_x = np.average(x)
            mean_y = np.average(y)
            sd_x = np.sum((x - mean_x)**2)**(1/2)
            sd_y = np.sum((y - mean_y)**2)**(1/2)
            prod_array = ((x - mean_x) * (y - mean_y)) / (sd_x * sd_y)

            correlation = np.sum(prod_array)
        else:
            number = 0
            correlation = 0

            for (result_list, result_shifted_list) in shift_list:
                for result in result_list:
                    for result_shifted in result_shifted_list:
                        correlation += result * result_shifted
                        number += 1

            if number >= 1:
                correlation /= number
            else:
                correlation = np.nan


        logger.debug('Correlation %f calculated from %d measurements.' % (correlation, number))

        return (correlation, number)


    def _iterate_over_shift_in_direction(self, calculate_function, direction, same_bounds, dim_ranges, wrap_around_t=False, file=None):
        logger.debug('Applying function to shifts by direction %s with same_bounds %s and dim_ranges %s.' % (direction, same_bounds, dim_ranges))

        ## init
        function_results_list = []
        direction_array = np.array(direction, dtype=np.float)
        if wrap_around_t:
            t_range = (dim_ranges[0][0], dim_ranges[0][1])
        else:
            t_range = None
        x_range = (dim_ranges[1][0], dim_ranges[1][1])

        ## calculate max factor
        if not np.all(direction == 0):
            dim_ranges_array = np.array(dim_ranges, dtype=np.float)
            dim_ranges_diff = dim_ranges_array[:,1] - dim_ranges_array[:,0]
            if wrap_around_t:
                dim_ranges_diff[0] = dim_ranges_diff[0] / 2
            dim_ranges_diff[1] = dim_ranges_diff[1] / 2
            max_factor_mask = direction_array != 0
            max_factor = math.floor(min(dim_ranges_diff[max_factor_mask] / direction_array[max_factor_mask]))
        else:
            max_factor = 0

        logger.debug('Max factor is %d.', max_factor)

        ## iterate over all factors
        for factor in range(max_factor + 1):
            shift_list = self._get_results_together_with_shifted(factor, direction, same_bounds, x_range, t_range)

            ## apply calculate_function to shift list
            logger.debug('Applying calculate function to shifts.')
            function_result = calculate_function(shift_list)
            function_results_list.append(function_result)

            ## save intermediate result
            if file is not None:
                function_results_array = np.array(function_results_list)
                np.save(file, function_results_array)

        ## create array and save results
        function_results_array = np.array(function_results_list)

        if file is None:
            np.save(file, function_results_array)

        logger.debug('Results array calculated.')

        return function_results_array


    def total_correlogram(self, direction, same_bounds, dim_ranges, wrap_around_t=False, minimum_measurements=1, is_normalized=False, file=None):

        logger.debug('Calculating correlogram.')

        calculate_correlation = lambda shift_list : self._calculate_total_correlation_from_shift_list(shift_list, is_normalized=is_normalized)
        correlogram = self._iterate_over_shift_in_direction(calculate_correlation, direction, same_bounds, dim_ranges, wrap_around_t=wrap_around_t, file=file)

        logger.debug('Correlogram calculated.')

        return correlogram


    def _iterate_over_shift_all_factor_combinations(self, calculation_function, direction, factor_lists, same_bounds, wrap_around_ranges, minimum_measurements=1, file=None):

        logger.debug('Iterate over all shifts with all factor combinations with the following configurations: direction=%s, factor_lists=%s, same_bounds=%s, wrap_around_ranges=%s, minimum_measurements=%d.' % (direction, factor_lists, same_bounds, wrap_around_ranges, minimum_measurements))

        function_results_list = []

        dim = len(direction)
        max_index = np.array([len(factor_list) for factor_list in factor_lists])

        measurements_dict_list = [[(self.measurements_dict, self.measurements_dict),],] + [None] * dim
        current_dim = 0
        current_indices = np.zeros(dim, dtype=np.int)
        current_shift = np.zeros(dim)

        ## iterate over all factors
        while current_dim >= 0:

            ## iterate over all dimensions
            while 0 <= current_dim < dim:
                current_index = current_indices[current_dim]
                current_factor_list = factor_lists[current_dim]

                ## search matching shifts
                current_factor = current_factor_list[current_index]
                current_shift[current_dim] = direction[current_dim] * current_factor
                measurements_dict_list[current_dim + 1] = self._get_first_dim_shifted(measurements_dict_list[current_dim], current_shift[current_dim], same_bounds[current_dim], wrap_around_ranges[current_dim])

                ## increase current dim
                current_dim += 1


            ## calculate value and append to list
            (value, number_of_measurements) = calculation_function(measurements_dict_list[dim])
            if number_of_measurements >= minimum_measurements:
                function_results_list.append(current_shift.tolist() + [value, number_of_measurements])
                logger.debug('Value %f for shift %s calculated and inserted. %d matching measurements where found.' % (value, current_shift, number_of_measurements))

                ## save intermediate result
                if file is not None:
                    function_results_array = np.array(function_results_list)
                    np.save(file, function_results_array)
            else:
                logger.debug('Value for shift %s not inserted. Only %d matching measurements where found.' % (current_shift, number_of_measurements))


            ## increase index
            current_dim -= 1
            measurements_dict_list[current_dim + 1] = None
            current_indices[current_dim] += 1

            while current_dim >= 0 and current_indices[current_dim] == len(factor_lists[current_dim]):
                measurements_dict_list[current_dim] = None
                current_indices[current_dim] = 0
                current_dim -= 1
                if current_dim >= 0:
                    current_indices[current_dim] += 1

        ## create array and save results
        function_results_array = np.array(function_results_list)

        if file is None:
            np.save(file, function_results_array)

        logger.debug('Iterate over all shifts with all factor combinations finished.')

        return function_results_array


    def total_correlation(self, direction, factor_lists, same_bounds, wrap_around_ranges, minimum_measurements=1, is_normalized=False, file=None):
        logger.debug('Calculating correlation with the following configurations: direction=%s, factor_lists=%s, same_bounds=%s, wrap_around_ranges=%s, minimum_measurements=%d.' % (direction, factor_lists, same_bounds, wrap_around_ranges, minimum_measurements))

        calculate_correlation = lambda shift_list : self._calculate_total_correlation_from_shift_list(shift_list, is_normalized=is_normalized)
        correlation = self._iterate_over_shift_all_factor_combinations(calculate_correlation, direction, factor_lists, same_bounds, wrap_around_ranges, minimum_measurements=minimum_measurements, file=file)

        logger.debug('Correlation calculated.')

        return correlation



class MeasurementsSamePointsDict(MeasurementsDict):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)


    ## compute values

    def correlation_or_covariance(self, value_type, min_number_of_values=10, stationary=False, max_year_diff=float('inf')):
        logger.debug('Calculate {} with at least {} values, stationary {} and max_year_diff {}.'.format(value_type, min_number_of_values, stationary, max_year_diff))

        ## check value type
        POSSIBLE_VALUE_TYPES = ('correlation', 'covariance')
        if value_type not in POSSIBLE_VALUE_TYPES:
            raise ValueError('Value type has to be in {} but it is {}.'.format(POSSIBLE_VALUE_TYPES, value_type))
        else:
            calculate_correlation = value_type == POSSIBLE_VALUE_TYPES[0]

        ## prepare value measurement dict
        if stationary:
            value_measurements = MeasurementsCovarianceStationaryDict()
        else:
            value_measurements = MeasurementsCovarianceDict()

        ## calculate values
        if len(self) > 0:

            ## check max_year_diff
            if max_year_diff is None or max_year_diff == float('inf'):
                t = self.values()[:,0]
                max_year_diff = int(np.ceil((t.max() - t.min()) / self.year_len))
                logger.debug('Using max_year_diff {}.'.format(max_year_diff))

            ## iterate over each pair of measurement indices
            index_of_measurement = [0, 0]
            for (key_0, transformed_value_list_0) in self.iterator_keys_and_value_lists():
                index_of_measurement[1] = 0
                for (key_1, transformed_value_list_1) in self.iterator_keys_and_value_lists():

                    ## skip previous values
                    if index_of_measurement[1] > index_of_measurement[0] or (index_of_measurement[1] == index_of_measurement[0] and max_year_diff > 1):

                        ## make keys to list
                        keys = (tuple(key_0), tuple(key_1))

                        ## calculate all desired year offsets
                        desired_year_offsets = tuple(np.arange(1, max_year_diff) * self.year_len)
                        if not np.allclose(keys[0], keys[1]):
                            desired_year_offsets += tuple(np.arange(-max_year_diff+1, 1) * self.year_len)

                        if not np.isclose(keys[0][0], keys[1][0]):
                            if keys[0][0] < keys[1][0]:
                                desired_year_offsets +=  (max_year_diff * self.year_len,)
                            else:
                                desired_year_offsets +=  (-max_year_diff * self.year_len,)

                        ## for all year offsets
                        for desired_year_offset in desired_year_offsets:

                            ## get values with desired year offset and t fraction diff
                            matching_results = []
                            for (t0, r0) in transformed_value_list_0:
                                for (t1, r1) in transformed_value_list_1:
                                    year_offset = (int(np.floor(t0 / self.year_len)) - int(np.floor(t1 / self.year_len))) * self.year_len
                                    if year_offset == desired_year_offset:
                                        matching_results.append([r0, r1])
                            matching_results = np.array(matching_results)
                            n = len(matching_results)


                            ## if enough measurements
                            if n >= min_number_of_values:

                                ## calculate auxiliary values
                                x0 = matching_results[:,0]
                                x1 = matching_results[:,1]

                                m0 = x0.mean()
                                m1 = x1.mean()

                                if calculate_correlation:
                                    s0 = np.sqrt(np.sum((x0 - m0)**2))
                                    s1 = np.sqrt(np.sum((x1 - m1)**2))

                                if not calculate_correlation or (s0 > 0 and s1 > 0):
                                    ## calculate value
                                    if calculate_correlation:
                                        value = np.sum(((x0 - m0) / s0) * ((x1 - m1) / s1))
                                        assert value >= -1 and value <= 1
                                    else:
                                        value = np.sum((x0 - m0) * (x1 - m1)) / (n-1)

                                    value = (n, value)

                                    ## prepare key pair
                                    value_keys = [list(keys[0]), list(keys[1])]
                                    if desired_year_offset >= 0:
                                        value_keys[0][0] = value_keys[0][0] + desired_year_offset
                                    else:
                                        value_keys[1][0] = value_keys[1][0] - desired_year_offset

                                    ## add value to value dict
                                    assert not value_measurements.has_values(value_keys)
                                    value_measurements.append_value(value_keys, value)

                                    logger.debug('{} {} calculated with {} values for index {}.'.format(value_type, value, n, value_keys))
                                else:
                                    logger.warning('Correlation for key {} and {} could not be calculated since a sample standard deviation is zero. Skipping this correlation.'.format(keys[0], keys[1]))


                    index_of_measurement[1] += 1
                index_of_measurement[0] += 1

        return value_measurements


    def correlation(self, min_number_of_values=10, stationary=False):
        return self.correlation_or_covariance('correlation', min_number_of_values=min_number_of_values, stationary=stationary)


    def covariance(self, min_number_of_values=10, stationary=False):
        return self.correlation_or_covariance('covariance', min_number_of_values=min_number_of_values, stationary=stationary)



class MeasurementsCovarianceDict(util.multi_dict.MultiDictPermutablePointPairs):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)
        self._year_len = 1


    def _preprocess_keys(self, keys):
        ## copy keys
        keys = list(keys)
        keys[0] = list(keys[0])
        keys[1] = list(keys[1])

        ## remove lower year offset
        year_len = self.year_len
        lower_years = min([int(keys[0][0]/year_len), int(keys[1][0]/year_len)])
        for key in keys:
            key[0] = key[0] - lower_years * year_len

        ## get value
        return super()._preprocess_keys(keys)


    @property
    def year_len(self):
        return self._year_len


    ## transform keys

    def coordinates_to_map_indices(self, lsm, int_indices=True):
        logger.debug('Transforming in {} coordinates to map indices of {} with int_indices {}'.format(self, lsm, int_indices))
        self._year_len = lsm.t_dim
        self.transform_keys(lambda keys: (lsm.coordinate_to_map_index(*keys[0], discard_year=False, int_indices=int_indices), lsm.coordinate_to_map_index(*keys[1], discard_year=False, int_indices=int_indices)))


    def map_indices_to_coordinates(self, lsm):
        logger.debug('Transforming in {} map indices of {} to coordinates'.format(self, lsm))
        self._year_len = 1
        self.transform_keys(lambda indices: (lsm.map_index_to_coordinate(*indices[0]), lsm.map_index_to_coordinate(*indices[1])))



    ## io
    #@overrides.overrides
    def save(self, file):
        only_dict = self.year_len != 1
        super().save(file, only_dict=only_dict)



class MeasurementsCovarianceStationaryDict(util.multi_dict.MultiDictDiffPointPairs, MeasurementsDict):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)





import os.path

from measurements.all.constants import BASE_DIR

WOA_BASE_DIR = os.path.join(BASE_DIR, 'woa')

NOBS_FILE =    os.path.join(WOA_BASE_DIR, 'dop_po4_2.8x2.8_monthly_nobs.npy')
VARIS_FILE =   os.path.join(WOA_BASE_DIR, 'dop_po4_2.8x2.8_monthly_vari.npy')
MEANS_FILE =   os.path.join(WOA_BASE_DIR, 'dop_po4_2.8x2.8_monthly_mean.npy')


import numpy as np


def average_over_time(data):
    averaged_data = np.nansum(data, axis=0) / np.sum(np.logical_not(np.isnan(data)), axis=0)
    return averaged_data


def wrap_around_index(index, index_range):
    if index_range is not None:
        index_range_diff = index_range[1] - index_range[0]
        while index < index_range[0]:
            index += index_range_diff
        while index >= index_range[1]:
            index -= index_range_diff
    return index


def normalize_points(points, ranges):
    points = np.asanyarray(points)
    ranges = np.asanyarray(ranges)
    return (points - range[:,0]) / (range[:,1] - range[:,0])


def get_min_distance(point_1, point_2, t_range=None, x_range=None):
    distance = np.abs(point_1 - point_2)

    ## wrap around
    if t_range is not None:
        t_size = t_range[1] - t_range[0]
        if distance[0] > t_size / 2:
            distance[0] -= t_size / 2

    if x_range is not None:
        x_size = x_range[1] - x_range[0]
        if distance[1] > x_size / 2:
            distance[1] -= x_size / 2

    return distance




import os.path

from measurements.constants import BASE_DIR

## base dir
WOA_BASE_DIR = os.path.join(BASE_DIR, 'po4/woa13')

## netcdf files
ANNUAL_FILE = os.path.join(WOA_BASE_DIR, 'data/annual/woa13_all_p00_01.nc')

MONTHLY_NUMBER_RANGE = range(1, 13)
MONTHLY_FILE_PATTERN = os.path.join(WOA_BASE_DIR, 'data/monthly/woa13_all_p{:0>2}_01.nc')
MONTHLY_FILES = []
for i in MONTHLY_NUMBER_RANGE:
    MONTHLY_FILES.append(MONTHLY_FILE_PATTERN.format(i))

## numpy files
MEANS_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_means.npy')
NOBS_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_nobs.npy')
VARIANCES_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_variances.npy')

## data names
MEAN_DATANAME = 'p_mn'
NOBS_DATANAME = 'p_dd'
STANDARD_DEVIATION_DATANAME = 'p_sd'
DEPTH_DATANAME = 'depth'

## ranges
T_RANGE = [0, 11]
X_RANGE = [0, 360]
Y_RANGE = [0, 180]

## variance interpolator
VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR = 1
VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR = 0
VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND = (0.1, 0.1, 0, 0)

## annual values
ANNUAL_THRESHOLD = 525


import numpy as np
import logging
logger = logging.getLogger(__name__)

import measurements.dop.pw.data
import measurements.util.regrid
import util.io.np



def save(land_sea_mask, z_values, t_dim=12):
    from measurements.dop.constants import DEVIATION_MIN_MEASUREMENTS, DEVIATION_MIN_VALUE, T_RANGE, X_RANGE, Y_RANGE
    from .constants import NOBS_FILE, VARIS_FILE, MEANS_FILE

    logger.debug('Calculating and saving dop measurement data.')

    ## load measurement data
    measurement_data = measurements.dop.pw.data.data()


    ## regrid data
    (means, nobs, variances) = measurements.util.regrid.measurements_to_land_sea_mask(measurement_data, land_sea_mask, t_range=T_RANGE, x_range=X_RANGE, y_range=Y_RANGE)


    ## remove variances with to few measurements
    variances[nobs < DEVIATION_MIN_MEASUREMENTS] = np.inf

    ## set variances below lower bound to lower bound
    variances[variances < DEVIATION_MIN_VALUE**2] = DEVIATION_MIN_VALUE**2

    ## where no variances use averaged variance
    variances_averaged = variances[np.isfinite(variances)].mean()
    variances[variances == np.inf] = variances_averaged


    ## save values
    util.io.np.save(MEANS_FILE, means, make_read_only=True, create_path_if_not_exists=True)
    util.io.np.save(NOBS_FILE, nobs, make_read_only=True, create_path_if_not_exists=True)
    util.io.np.save(VARIS_FILE, variances, make_read_only=True, create_path_if_not_exists=True)


import bisect
import os

import numpy as np
#import overrides

import util.cache.file
import util.cache.memory
import util.petsc.universal
import util.logging

import measurements.constants
import measurements.land_sea_mask.depth
import measurements.land_sea_mask.constants

logger = util.logging.logger



class LandSeaMask():

    def __init__(self, lsm, depth_level, t_dim=None, t_centered=True):
        assert lsm.ndim == 2
        assert len(depth_level) > 0

        self._lsm = lsm
        self._depth_level = np.asanyarray(depth_level)

        self._t_dim = t_dim
        self.t_centered = t_centered


    ## equal

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.dim == other.dim and self.t_centered == other.t_centered and np.all(self.z == other.z) and np.all(self.lsm == other.lsm)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented


    ## dims

    @property
    def t_dim(self):
        return self._t_dim

    @t_dim.setter
    def t_dim(self, new_t_dim):
        self._t_dim = new_t_dim

    @property
    def t_dim_with_value_check(self):
        t_dim = self.t_dim
        if t_dim is not None and t_dim != 0:
            return t_dim
        else:
            raise ValueError('T dim is not set.')
        self.sea_indices.cache_clear()
        self.sea_coordinates.cache_clear()


    @property
    def x_dim(self):
        return self._lsm.shape[0]

    @property
    def y_dim(self):
        return self._lsm.shape[1]

    @property
    def z_dim(self):
        return len(self._depth_level) - 1

    @property
    def space_dim(self):
        return (self.x_dim, self.y_dim, self.z_dim)

    @property
    def dim(self):
        if self.t_dim is None or self.t_dim == 0:
            return (self.x_dim, self.y_dim, self.z_dim)
        else:
            return (self.t_dim, self.x_dim, self.y_dim, self.z_dim)

    @property
    def ndim(self):
        return len(self.dim)


    ## z

    @property
    def z(self):
        return self._depth_level

    @z.setter
    def z(self, new_z_values):
        logger.debug('Regridding z from {} to {}.'.format(self.z, new_z_values))
        new_z_values = np.asanyarray(new_z_values)

        old_z_values = self.z
        for i in range(len(old_z_values)):
            self._lsm[self._lsm == i] = bisect.bisect_left(new_z_values, old_z_values[i])
        self._depth_level = new_z_values

    @property
    def z_left(self):
        return self.z[:-1]

    @property
    def z_right(self):
        return self.z[1:]

    @property
    def z_center(self):
        return (self.z_left + self.z_right) / 2

    @property
    def separation_values(self):
        return (1/self.t_dim_with_value_check, 360/self.x_dim, 180/self.y_dim, self.z)


    ## lsm

    @property
    def lsm(self):
        return self._lsm

    @property
    def shape(self):
        return self._lsm.shape

    def __getitem__(self, key):
        if len(key) == 2:
            return self._lsm[key]
        elif len(key) == 3:
            return self._lsm[key[1:]]
        elif len(key) == 4:
            return self._lsm[key[1:3]] > key[3]
        else:
            raise ValueError('Length of key has to be in (2, 3, 4), but key is {}.'.format(key))


    def __str__(self):
        try:
            t_dim = self.t_dim_with_value_check
        except ValueError:
            t_dim = None
        if t_dim is not None:
            return 'lsm_{}'.format(t_dim)
        else:
            return 'lsm'


    ## indices

    @property
    @util.cache.memory.method_decorator(dependency='self.t_dim')
    def sea_indices(self):
        sea_indices = np.array(np.where(self.bool_mask)).transpose()
        logger.debug('Found {} sea indices in {}.'.format(sea_indices.shape[0], self))
        assert sea_indices.ndim == 2
        return sea_indices


    @property
    @util.cache.memory.method_decorator(dependency='self.t_dim')
    def sea_coordinates(self):
        sea_coordinates = self.map_indices_to_coordinates(self.sea_indices)
        assert sea_coordinates.ndim == 2
        return sea_coordinates


    def is_coordinate_near_water(self, point, max_box_distance_to_water=0):
        if max_box_distance_to_water < 0:
            raise ValueError('max_box_distance_to_water must be greater or equal to 0 but it is {}.'.format(max_box_distance_to_water))
        old_t_dim =  self.t_dim
        self.t_dim = 0

        sea_indices = self.sea_indices
        map_index = np.asarray(self.coordinate_to_map_index(*point, discard_year=True, int_indices=True))
        distance = np.abs(sea_indices - map_index[np.newaxis, :])

        self.t_dim = old_t_dim

        is_near_water = np.any(np.all(distance <= max_box_distance_to_water, axis=1))
        logger.debug('Coordinate {} is near water {} with max_box_distance_to_water {}.'.format(point, is_near_water, max_box_distance_to_water))
        return is_near_water


    def coordinates_near_water_mask(self, points, max_box_distance_to_water=0):
        n = len(points)
        results = np.empty(n, dtype=np.bool)
        for i in range(n):
            results[i] = self.is_coordinate_near_water(points[i], max_box_distance_to_water=max_box_distance_to_water)
        return results


    def box_bounds_of_map_index(self, map_index):
        map_index = np.asarray(map_index)
        t_centered = self.t_centered and len(map_index) >= 4

        map_index = map_index + 0.5
        if not t_centered:
            map_index[0] = map_index[0] -  0.5

        lower_bound = np.floor(map_index)
        lower_bound = lower_bound - 0.5
        if not t_centered:
            lower_bound[0] = lower_bound[0] +  0.5

        box_bounds = np.array([lower_bound, lower_bound + 1])
        box_bounds = box_bounds.transpose()

        assert len(box_bounds) == len(map_index)
        assert box_bounds.shape[1] == 2
        assert np.all(box_bounds[:,1] - box_bounds[:,0] == 1)
        return box_bounds


    def box_bounds_of_map_indices(self, map_indices):
        ## prepare input
        result_ndim = map_indices.ndim
        if map_indices.ndim == 1:
            map_indices = map_indices[np.newaxis]

        ## calculate
        n = len(map_indices)
        logger.debug('Transforming {} map indices to box bounds for {}.'.format(n, self))

        box_bounds = np.empty(map_indices.shape + (2,))
        for i in range(n):
            box_bounds[i] = self.box_bounds_of_map_index(map_indices[i])

        ## return
        if result_ndim == 1:
            box_bounds = box_bounds[0]

        logger.debug('Transforming map indices to box bounds done.')
        return box_bounds


    @property
    @util.cache.memory.method_decorator(dependency='self.t_dim')
    def number_of_map_indices(self):
        t_dim = self.t_dim
        if t_dim is None:
            t_dim = 1
        return self.lsm.sum() * self.t_dim


    ## volume

    @staticmethod
    def volume_of_coordinate_box(bounds):
        bounds = np.asanyarray(bounds)
        assert bounds.ndim == 2
        assert bounds.shape[0] >= 3
        assert bounds.shape[1] == 2
        # assert bounds.shape == (3, 2)

        bounds = bounds[-3:]
        assert np.all(bounds[1] >= -90)
        assert np.all(bounds[1] <= 90)
        assert np.all(bounds[2] >= 0)

        alpha = bounds[0]
        beta = bounds[1]
        r = measurements.constants.EARTH_RADIUS - bounds[2]
        r = r[::-1]

        s = np.pi / (2 * 90)
        v = 1/3 * s * (alpha[1] - alpha[0]) *  (r[1]**3 - r[0]**3) * (np.sin(s * beta[1]) - np.sin(s * beta[0]))
        assert v >= 0
        return v


    @staticmethod
    def volume_of_coordinate_boxes(bounds):
        ## prepare input
        bounds = np.asanyarray(bounds)

        result_ndim = bounds.ndim
        if bounds.ndim == 2:
            bounds = bounds[np.newaxis]

        assert bounds.ndim == 3
        assert bounds.shape[1] >= 3
        assert bounds.shape[2] == 2

        ## calculate
        n = len(bounds)
        logger.debug('Calculating volume of {} coordinate boxes.'.format(n))

        box_volumes = np.empty(n)
        for i in range(n):
            box_volumes[i] = LandSeaMask.volume_of_coordinate_box(bounds[i])

        ## return
        if result_ndim == 1:
            box_volumes = box_volumes[0]

        logger.debug('Volume of coordinate boxes are calculated.')
        return box_volumes


    def volume_of_boxes_of_map_indices(self, map_indices):
        logger.debug('Calculating volume of boxes of {} map indices.'.format(len(map_indices)))

        ## calculate box bounds as map indices
        box_bounds = self.box_bounds_of_map_indices(map_indices)
        assert box_bounds.shape[2] == 2
        assert np.all(box_bounds[:,:,1] > box_bounds[:,:,0])

        ## calculate box bounds as coordinates
        for i in range(box_bounds.shape[2]):
           box_bounds[:,:,i] = self.map_indices_to_coordinates(box_bounds[:,:,i], use_modulo_for_x=False)
        assert np.all(box_bounds[:,:,1] > box_bounds[:,:,0])

        ## calculate volumes
        volumes = self.volume_of_coordinate_boxes(box_bounds)

        ## return
        return volumes


    @property
    def volume_map(self):
        logger.debug('Calculating volume map.')

        ## calculate sea map indices
        sea_indices = self.sea_indices

        ## calculate volume map
        volumes = self.volume_of_boxes_of_map_indices(sea_indices)
        volumes_with_indices = np.concatenate([sea_indices, volumes[:,np.newaxis]], axis=1)
        volume_map = self.insert_index_values_in_map(volumes_with_indices, no_data_value=np.inf)

        ## return
        return volume_map


    @property
    def normalized_volume_weights_map(self):
        volume_map = self.volume_map
        normalized_volume_map = volume_map / np.nansum(volume_map)
        assert np.isclose(np.nansum(normalized_volume_map), 1)
        return normalized_volume_map


    ## convert map indices and coordinates

    def coordinate_to_map_index(self, t, x, y, z, discard_year=False, int_indices=True):
        ## t (center of the box, wrap around)
        try:
            t_dim = self.t_dim_with_value_check
        except ValueError:
            t_dim = None

        if t_dim is not None:
            ti = t * t_dim
            if discard_year:
                ti = (ti % t_dim)
            if self.t_centered:
                ti -= 0.5

        ## x (center of the box, wrap around)
        xi = (x % 360) / 360 * self.x_dim - 0.5

        ## y (center of the box, no wrap around)
        yi = (y + 90) / 180 * self.y_dim  - 0.5

        ## z (center of the box, no wrap around)
        r = self.z_right
        c = self.z_center
        m = len(c)-1

        zi = bisect.bisect_left(c, z) - 1
        if zi == -1:
            assert z <= c[0]
            zi = 1/2 * (z / c[0] - 1)
        elif zi == m:
            assert z >= c[m]
            zi = m +  1/2 * (z - c[m]) / (r[m] - c[m])
        else:
            assert z >= c[zi] and z <= c[zi+1]
            zi += 1/2 * (min(z, r[zi]) - c[zi]) / (r[zi] - c[zi]) + 1/2 * (max(z, r[zi]) - r[zi]) / (c[zi+1] - r[zi])

        ## concatenate (float) index
        if t_dim is not None:
            map_index = (ti, xi, yi, zi)
        else:
            map_index = (xi, yi, zi)

        ## convert to int if needed
        if int_indices:
            ## round half up (and not round half to even which is numpys default)
            map_index = np.array(map_index)
            tie_mask = map_index % 1 == 0.5
            map_index[tie_mask] = map_index[tie_mask] + 0.5
            map_index = np.array(np.round(map_index), dtype=np.int32)
            ## check bounds
            if map_index[-1] > self.z_dim:          # below z bottom
                map_index[-1] = self.z_dim
            if map_index[-2] == self.y_dim:         # y = 90 degree
                map_index[-2] = self.y_dim - 1
            ## convert back to tuple
            map_index = tuple(map_index)
            assert len(map_index) == 3 or (not discard_year) or (map_index[0] >= 0 and map_index[0] < self.t_dim)
            assert (map_index[-3] >= 0 and map_index[-3] < self.x_dim)
            assert (map_index[-2] >= 0 and map_index[-2] < self.y_dim)
            assert (map_index[-1] >= 0 and map_index[-1] <= self.z_dim)

        ## return
        return map_index


    def coordinates_to_map_indices(self, points, discard_year=False, int_indices=True):
        points = np.asanyarray(points)
        result_ndim = points.ndim
        if points.ndim == 1:
            points = points[np.newaxis]
        logger.debug('Transforming {} coordinates to map indices for {} with discard year {} and int_indices {}.'.format(len(points), self, discard_year, int_indices))

        n = len(points)
        if int_indices:
            dtype= np.int32
        else:
            dtype= np.float
        new_points = np.empty((n, self.ndim), dtype=dtype)

        for i in range(n):
            new_points[i] = self.coordinate_to_map_index(*points[i], discard_year=discard_year, int_indices=int_indices)

        if result_ndim == 1:
            new_points = new_points[0]

        logger.debug('Transforming from coordinates to map indices done.')
        return new_points


    def map_index_to_coordinate(self, ti, xi, yi, zi, use_modulo_for_x=True):
        ## t (left or center of the box, wrap around)
        try:
            t_dim = self.t_dim_with_value_check
        except ValueError:
            t_dim = None
        if t_dim is not None:
            if self.t_centered:
                ti += 0.5
            t = ti / self.t_dim_with_value_check

        ## x (center of the box, wrap around)
        x = xi + 0.5
        if use_modulo_for_x:
            x = x % self.x_dim
        x = x / self.x_dim * 360

        ## y (center of the box, no wrap around)
        y = (yi + 0.5) / self.y_dim * 180 - 90

        ## z (center of the box, no wrap around)
        r = self.z_right
        c = self.z_center
        m = len(c)-1

        if zi < 0:
            z = (2 * zi + 1) * c[0]
            assert z <= c[0]
        elif zi >= m:
            z = 2 * (zi - m) * (r[m] - c[m]) + c[m]
            assert z >= c[m]
        else:
            zi_floor = int(np.floor(zi))
            zi_fraction = zi % 1
            if zi_fraction < 0.5:
                z = 2 * zi_fraction * (r[zi_floor] - c[zi_floor]) + c[zi_floor]
            else:
                zi_fraction -= 0.5
                z = 2 * zi_fraction * (c[zi_floor + 1] - r[zi_floor]) + r[zi_floor]
            assert z >= c[zi_floor] and z <= c[zi_floor+1]

        ## concatenate coordinates
        if t_dim is not None:
            coordinates = (t, x, y, z)
        else:
            coordinates = (x, y, z)

        ## return
        assert not use_modulo_for_x or (coordinates[-3] >= 0 and coordinates[-3] <= 360)
        assert coordinates[-2] >= -90
        assert coordinates[-2] <= 90
        assert coordinates[-1] >= 0
        assert coordinates[-1] <= measurements.constants.MAX_SEA_DEPTH
        return coordinates


    def map_indices_to_coordinates(self, points, use_modulo_for_x=True):
        result_ndim = points.ndim
        if points.ndim == 1:
            points = points[np.newaxis]
        logger.debug('Transforming {} map indices from {} to coordinates with use_modulo_for_x {}'.format(len(points), self, use_modulo_for_x))

        n = len(points)
        new_points = np.empty((n, self.ndim))
        for i in range(n):
            new_points[i] = self.map_index_to_coordinate(*points[i], use_modulo_for_x=use_modulo_for_x)

        if result_ndim == 1:
            new_points = new_points[0]

        logger.debug('Transforming from map indices to coordinates done.')
        return new_points


    ## values to map

    def apply_mask(self, array, land_value=np.nan):
        if self.dim != array.shape:
            raise ValueError('Array must have the same dims as lsm, but its shape is {} and it has to be {}.'.format(array.shape, self.dim))

        for i in np.ndindex(self.dim[:-1]):
            z_max = self[i]
            array[i][z_max:] = land_value
        return array


    def masked_map(self, default_value=0, land_value=np.nan, dtype=np.float64):
        masked_map = np.ones(self.dim, dtype=dtype) * default_value
        self.apply_mask(masked_map, land_value=land_value)
        return masked_map


    @property
    def bool_mask(self):
        return self.masked_map(dtype=np.bool, default_value=True, land_value=False)


    def insert_coordinate_values_in_map(self, values, no_data_value=0, skip_values_on_land=True):
        values = np.copy(values)
        values[:,:-1] = self.coordinates_to_map_indices(values[:,:-1], discard_year=True, int_indices=True)
        return self.insert_index_values_in_map(values, no_data_value=no_data_value, skip_values_on_land=skip_values_on_land)


    def insert_index_values_in_map(self, values, no_data_value=0, skip_values_on_land=True):
        logger.debug('Inserting {} values in map with value {} for no data.'.format(len(values), no_data_value))

        if values.shape[1] not in (4, 5):
            raise ValueError('Values have wrong shape: Second dimension have to be 4 or 5, but it is {}.'.format(values.shape[1]))
        if np.isnan(no_data_value):
            raise ValueError('No data value can not be NAN.')

        ## remove time dim if values have no time
        if values.shape[1] == 4:
            old_t_dim = self.t_dim
            self.t_dim = None

        ## init map
        value_map = self.masked_map(default_value=no_data_value, land_value=np.nan, dtype=values.dtype)
        number_map = self.masked_map(default_value=0, land_value=0, dtype=np.int32)

        ## insert values: sum and count for each box
        for row in values:
            index = tuple(row[:-1].astype(np.int))
            value = row[-1]
            try:
                value_at_index = value_map[index]
            except IndexError:
                raise ValueError('Index {} exceeds dimension {}.'.format(index, value_map.shape))

            if not skip_values_on_land or not np.isnan(value_at_index):
                if value_at_index == no_data_value or np.isnan(value_at_index):
                    value_map[index] = value
                else:
                    value_map[index] = value_at_index + value
                number_map[index] = number_map[index] + 1

        ## average
        mask = number_map > 1
        value_map[mask] = value_map[mask] / number_map[mask]

        ## restore time dim
        if values.shape[1] == 4:
            self.t_dim = old_t_dim

        ## return
        return value_map


    ## plot
    def plot(self):
        import util.plot
        file = '/tmp/{}.png'.format(self)
        util.plot.data(self.lsm, file, land_value=0, power_limit=10)


    ## copy
    def copy(self):
        import copy
        return copy.deepcopy(self)



class LandSeaMaskFromFile(LandSeaMask):

    def __init__(self, lsm_dir, t_dim=None, t_centered=True):
        self._lsm_file = os.path.join(lsm_dir, measurements.land_sea_mask.constants.LSM_NPY_FILENAME)
        self._depth_file = os.path.join(lsm_dir, measurements.land_sea_mask.constants.DEPTH_NPY_FILENAME)

        depth = self._calculate_depth()
        lsm = self._calculate_lsm()

        super().__init__(lsm, depth, t_dim=t_dim, t_centered=t_centered)


    @util.cache.file.decorator(cache_file_function=lambda self: self._lsm_file)
    def _calculate_lsm(self):
        raise NotImplementedError

    @util.cache.file.decorator(cache_file_function=lambda self: self._depth_file)
    def _calculate_depth(self):
        raise NotImplementedError



class LandSeaMaskTMM(LandSeaMaskFromFile):

    def __init__(self, t_dim=None, t_centered=True):
        super().__init__(measurements.land_sea_mask.constants.TMM_DIR, t_dim=t_dim, t_centered=t_centered)


    @util.cache.file.decorator(cache_file_function=lambda self: self._lsm_file)
    #@overrides.overrides
    def _calculate_lsm(self):
        lsm = util.petsc.universal.load_petsc_mat_to_array(measurements.land_sea_mask.constants.TMM_PETSC_FILE, dtype=np.int16)
        lsm = lsm.transpose() # metos3d: x and y are changed
        assert lsm.shape == (128, 64) and lsm.min() == 0 and lsm.max() == 15
        return lsm


    @util.cache.file.decorator(cache_file_function=lambda self: self._depth_file)
    #@overrides.overrides
    def _calculate_depth(self):
        ## read values from txt
        depth = np.genfromtxt(measurements.land_sea_mask.constants.TMM_DEPTH_TXT_FILE, dtype=np.int16, comments='#', usecols=(0,))
        assert depth.ndim == 1 and depth.shape[0] == 16
        return depth


    def __str__(self):
        return super().__str__() + '_tmm'



class LandSeaMaskWOA13(LandSeaMaskFromFile):

    def __init__(self, t_dim=None, t_centered=True):
        super().__init__(measurements.land_sea_mask.constants.WOA13_DIR, t_dim=t_dim, t_centered=t_centered)


    @util.cache.file.decorator(cache_file_function=lambda self: self._lsm_file)
    #@overrides.overrides
    def _calculate_lsm(self):
        ## read values from txt with axis order: x y z
        lsm = np.genfromtxt(measurements.land_sea_mask.constants.WOA13_LSM_TXT_FILE, dtype=float, delimiter=',', comments='#', usecols=(1, 0, 2))

        ## normalize values
        lsm[:,0] = lsm[:,0] % 360

        lsm = lsm - lsm.min(axis=0)

        ## convert to int
        lsm_int = lsm.astype(np.int16)

        assert np.all(lsm_int == lsm)
        assert lsm_int[:,0].min() == 0 and lsm_int[:,0].max() == 359 and lsm_int[:,1].min() == 0 and lsm_int[:,1].max() == 179 and lsm_int[:,2].min() == 0 and lsm_int[:,2].max() == 137

        ## convert in 2 dim
        lsm = np.empty((360, 180), dtype=np.int16)
        for x, y, z in lsm_int:
            lsm[x, y] = z

        assert lsm.min() == 0 and lsm.max() == 137
        return lsm


    @util.cache.file.decorator(cache_file_function=lambda self: self._depth_file)
    #@overrides.overrides
    def _calculate_depth(self):
        ## read values from txt
        depth = np.genfromtxt(measurements.land_sea_mask.constants.WOA13_DEPTH_TXT_FILE, dtype=np.int16, comments='#', usecols=(0,))
        assert depth.ndim == 1 and depth.shape[0] == 138
        return depth


    def __str__(self):
        return super().__str__() + '_woa13'



class LandSeaMaskWOA13R(LandSeaMask):

    def __init__(self, t_dim=None, t_centered=True):
        depth = measurements.land_sea_mask.depth.values_TMM(max_value=5200, increment_step=2)
        depth.extend([6000, 8000, 10000])

        lsm_woa13 = LandSeaMaskWOA13()
        lsm_woa13.z = depth
        lsm = lsm_woa13.lsm

        super().__init__(lsm, depth, t_dim=t_dim, t_centered=t_centered)

    def __str__(self):
        return super().__str__() + '_woa13r'






import numpy as np

import measurements.land_sea_mask.constants
import util.petsc.universal

import util.logging
logger = util.logging.logger


def _check_land_sea_mask(land_sea_mask):
    ## check input
    if land_sea_mask.ndim != 2:
        raise ValueError('The land sea mask must have 2 dimensions, but its shape is {}.'.format(land_sea_mask.shape))


def resolution_128x64x15():
    from .constants import LSM_128x64x15_PETSC_FILE, LSM_128x64x15_NPY_FILE

    try:
        land_sea_mask = np.load(LSM_128x64x15_NPY_FILE)

        logger.debug('Returning land-sea-mask loaded from {} file.'.format(LSM_128x64x15_NPY_FILE))

    except (OSError, IOError):
        land_sea_mask = util.petsc.universal.load_petsc_mat_to_array(LSM_128x64x15_PETSC_FILE, dtype=int)
        land_sea_mask = land_sea_mask.transpose() # metos3d: x and y are changed

        _check_land_sea_mask(land_sea_mask)
        logger.debug('Saving land-sea-mask to {} file.'.format(LSM_128x64x15_NPY_FILE))
        np.save(LSM_128x64x15_NPY_FILE, land_sea_mask)
        logger.debug('Returning land-sea-mask loaded from petsc file.')


    return land_sea_mask


def resolution_128x64x15_normalized_volumes():
    from .constants import LSM_128x64x15_NORMALIZED_VLUMES_PETSC_FILE
    logger.debug('Loading normalized valumes from petsc file {}.'.format(LSM_128x64x15_NORMALIZED_VLUMES_PETSC_FILE))
    return util.petsc.universal.load_petsc_vec_to_numpy_array(LSM_128x64x15_NORMALIZED_VLUMES_PETSC_FILE)



## 60x180x138

def resolution_360x180x138():
    from .constants import LSM_360x180x138_TXT_FILE, LSM_360x180x138_NPY_FILE

    try:
        lsm = np.load(LSM_360x180x138_NPY_FILE)
        logger.debug('Returning land-sea-mask loaded from {} file.'.format(LSM_360x180x138_NPY_FILE))

    except (OSError, IOError):
        ## read values from txt with axis order: x y z
        lsm = np.genfromtxt(LSM_360x180x138_TXT_FILE, dtype=float, delimiter=',', comments='#', usecols=(1, 0, 2))

        ## normalize values
        min_values = lsm.min(axis=0)
        lsm[:,0] = lsm[:,0] - min_values[0]
        lsm[:,1] = lsm[:,1] - min_values[1]
        lsm[:,2] = lsm[:,2] - 1

        ## convert to int
        lsm_int = lsm.astype(np.int16)

        assert np.all(lsm_int == lsm)
        assert lsm_int[:,0].min() == 0 and lsm_int[:,0].max() == 359 and lsm_int[:,1].min() == 0 and lsm_int[:,1].max() == 179 and lsm_int[:,2].min() == 0 and lsm_int[:,2].max() == 137

        ## convert in 2 dim
        lsm = np.empty((360, 180), dtype=np.int16)
        for x, y, z in lsm_int:
            lsm[x, y] = z
        assert lsm.min() == 0 and lsm.max() == 137

        ## save lsm as npy
        _check_land_sea_mask(lsm)
        logger.debug('Saving land-sea-mask to {} file.'.format(LSM_360x180x138_NPY_FILE))
        np.save(LSM_360x180x138_NPY_FILE, lsm)
        logger.debug('Returning land-sea-mask loaded from txt file.')

    return lsm


def depth_360x180x138():
    from .constants import DEPTH_360x180x138_TXT_FILE, DEPTH_360x180x138_NPY_FILE

    ## read from cache
    try:
        depth = measurements.land_sea_mask.constants.DEPTH_360x180x138
        logger.debug('Returning depth levels from cache.')

    ## read from npy file
    except AttributeError:
        try:
            depth = np.load(DEPTH_360x180x138_NPY_FILE)
            logger.debug('Returning depth levels loaded from {} file.'.format(DEPTH_360x180x138_NPY_FILE))
            measurements.land_sea_mask.constants.DEPTH_360x180x138 = depth

    ## read from txt file
        except (OSError, IOError):
            ## read values from txt with axis order: x y z
            depth = np.genfromtxt(DEPTH_360x180x138_TXT_FILE, dtype=np.int16, comments='#', usecols=(0,))
            assert depth.ndim == 1 and depth.shape[0] == 138

            ## save depth as npy
            logger.debug('Saving depth to {} file.'.format(DEPTH_360x180x138_NPY_FILE))
            np.save(DEPTH_360x180x138_NPY_FILE, depth)
            logger.debug('Returning land-sea-mask loaded from txt file.')
            measurements.land_sea_mask.constants.DEPTH_360x180x138 = depth

    return depth



import calendar
import datetime
import os.path

import numpy as np
#import overrides

import util.math.sort
import util.cache.file
import util.cache.memory
import util.logging

import measurements.universal.data
import measurements.universal.constants
import measurements.constants
import measurements.dop.constants
import measurements.dop.pw.constants

logger = util.logging.logger


## data load functions

def prepare_ladolfi_data(measurement_file, start_date, end_date, valid_data_flag):

    # convert missing values to - Inf
    def convert_missing_values_to_nan(value_string):
        try:
            value = float(value_string)
        except ValueError:
            value = - float('nan')

        return value

    # load data, columns: lat, long, depth, dop, flag
    data = np.loadtxt(measurement_file, converters={3: convert_missing_values_to_nan})
    if data.shape[1] != 5:
        raise ValueError('The data in {} has not 5 columns. Its shape is {}'.format(measurement_file, data.shape))

    # skip invalid data and flag
    data = data[data[:,4]==valid_data_flag]
    data = data[:,:4]
    data = data[data[:,3]>=0]
    data = data[np.logical_not(np.isnan(data[:,3]))]

    # move columns in order long, lat, depth, dop
    data[:,[0, 1]] = data[:,[1, 0]]

    # sort data in priority by long, lat, depth
    data = data[np.lexsort((data[:,2], data[:,1], data[:,0]))]

    # interpolate time
    def convert_date_to_year_float(date):
        number_of_days_in_year = date.timetuple().tm_yday
        number_of_total_days_in_year = calendar.isleap(date.year) + 365
        year_frac = number_of_days_in_year / number_of_total_days_in_year
        year = date.year
        time = year + year_frac
        return time

    start_year_float = convert_date_to_year_float(start_date)
    end_year_float = convert_date_to_year_float(end_date)

    n = data.shape[0]
    t = np.arange(n) / (n-1) * (end_year_float - start_year_float) + start_year_float
    t = t.reshape([n, 1])
    data = np.concatenate((t, data), axis=1)

    logger.debug('{} DOP data sets loaded from {}.'.format(data.shape[0], measurement_file))

    return data


@util.cache.file.decorator(cache_file_function=lambda: os.path.join(measurements.dop.pw.constants.LADOLFI_2002_DIR, measurements.dop.pw.constants.DATA_FILENAME))
def load_ladolfi_2002():
    from measurements.dop.pw.constants import LADOLFI_2002_MEASUREMENT_FILE, LADOLFI_2002_START_DATE, LADOLFI_2002_END_DATE, LADOLFI_2002_VALID_DATA_FLAG
    return prepare_ladolfi_data(LADOLFI_2002_MEASUREMENT_FILE, LADOLFI_2002_START_DATE, LADOLFI_2002_END_DATE, LADOLFI_2002_VALID_DATA_FLAG)


@util.cache.file.decorator(cache_file_function=lambda: os.path.join(measurements.dop.pw.constants.LADOLFI_2004_DIR, measurements.dop.pw.constants.DATA_FILENAME))
def load_ladolfi_2004():
    from measurements.dop.pw.constants import LADOLFI_2004_MEASUREMENT_FILE, LADOLFI_2004_START_DATE, LADOLFI_2004_END_DATE, LADOLFI_2004_VALID_DATA_FLAG
    return prepare_ladolfi_data(LADOLFI_2004_MEASUREMENT_FILE, LADOLFI_2004_START_DATE, LADOLFI_2004_END_DATE, LADOLFI_2004_VALID_DATA_FLAG)



def prepare_yoshimura_data(measurement_file):
    ## columns: lat, long, date, time, dop

    ## calculate time
    def convert_date_to_year(value_bytes):
        value_string = value_bytes.decode()
        d = datetime.datetime.strptime(value_string, '%d-%B-%Y').date()
        year = d.year
        logger.debug('Convering: "{}" is in the year {}.'.format(value_bytes, year))
        return year

    def convert_date_to_number_of_days_in_year(value_bytes):
        value_string = value_bytes.decode()
        d = datetime.datetime.strptime(value_string, '%d-%B-%Y').date()
        number_of_days = d.timetuple().tm_yday
        logger.debug('Convering: "{}" is the {}. day in a year.'.format(value_bytes, number_of_days))
        return number_of_days

    def convert_date_to_number_of_all_days_in_year(value_bytes):
        value_string = value_bytes.decode()
        d = datetime.datetime.strptime(value_string, '%d-%B-%Y').date()
        number_of_days_in_year = calendar.isleap(d.year) + 365
        logger.debug('Convering: The year of "{}" has {} days.'.format(value_bytes, number_of_days_in_year))
        return number_of_days_in_year

    def convert_time_to_day_frac(value_bytes):
        value_string = value_bytes.decode()
        dt = datetime.datetime.strptime(value_string, '%H:%M')
        hour_frac = dt.hour + dt.minute / 60
        day_frac = hour_frac / 24
        logger.debug('Convering: The time "{}" is {} of a day.'.format(value_bytes, day_frac))
        return day_frac

    year = np.loadtxt(measurement_file, usecols=(2,), converters={2: convert_date_to_year})

    number_of_days_in_year = np.loadtxt(measurement_file, usecols=(2,), converters={2: convert_date_to_number_of_days_in_year})
    number_of_all_days_in_year = np.loadtxt(measurement_file, usecols=(2,), converters={2: convert_date_to_number_of_all_days_in_year})
    day_frac = np.loadtxt(measurement_file, usecols=(3,), converters={3: convert_time_to_day_frac})
    year_frac = (number_of_days_in_year + day_frac) / number_of_all_days_in_year

    time = year + year_frac


    ## load lat, long and dop
    lat = np.loadtxt(measurement_file, usecols=(0,))
    long = np.loadtxt(measurement_file, usecols=(1,))
    dop = np.loadtxt(measurement_file, usecols=(4,))
    depth = dop * 0


    ## concatenate columns in order long, lat, depth, dop
    data = np.concatenate((time[:,np.newaxis], long[:,np.newaxis], lat[:,np.newaxis], depth[:,np.newaxis], dop[:,np.newaxis]), axis=1)

    logger.debug('{} DOP data sets loaded from {}.'.format(data.shape[0], measurement_file))

    return data


@util.cache.file.decorator(cache_file_function=lambda: os.path.join(measurements.dop.pw.constants.YOSHIMURA_2007_DIR, measurements.dop.pw.constants.DATA_FILENAME))
def load_yoshimura_2007():
    from measurements.dop.pw.constants import YOSHIMURA_2007_MEASUREMENT_FILE
    return prepare_yoshimura_data(YOSHIMURA_2007_MEASUREMENT_FILE)



## measurement classes

class MeasurementsSingleBase(measurements.universal.data.MeasurementsAnnualPeriodicCache):

    def __init__(self, data_set_name, load_data_function, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):

        tracer = 'dop'
        data_set_name = data_set_name

        sample_lsm = measurements.dop.pw.constants.SAMPLE_LSM
        min_standard_deviation = measurements.dop.constants.DEVIATION_MIN_VALUE

        super().__init__(sample_lsm, tracer=tracer, data_set_name=data_set_name, min_standard_deviation=min_standard_deviation, min_measurements_correlation=min_measurements_correlation)

        self._load_data_function = load_data_function


    @property
    def points_and_results(self):
        values = self._load_data_function()
        sorted_indices = util.math.sort.lex_sorted_indices(values)
        values = values[sorted_indices]
        return values

    @property
    @util.cache.memory.method_decorator()
    @util.cache.file.decorator()
    #@overrides.overrides
    def points(self):
        return self.points_and_results[:, :-1]

    @property
    @util.cache.memory.method_decorator()
    @util.cache.file.decorator()
    #@overrides.overrides
    def values(self):
        return self.points_and_results[:, -1]




class MeasurementsLadolfi2002(MeasurementsSingleBase):

    def __init__(self, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        super().__init__('ladolfi_2002', load_ladolfi_2002, min_measurements_correlation=min_measurements_correlation)


class MeasurementsLadolfi2004(MeasurementsSingleBase):

    def __init__(self, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        super().__init__('ladolfi_2004', load_ladolfi_2004, min_measurements_correlation=min_measurements_correlation)


class MeasurementsYoshimura2007(MeasurementsSingleBase):

    def __init__(self, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        super().__init__('yoshimura_2007', load_yoshimura_2007, min_measurements_correlation=min_measurements_correlation)


class MeasurementsBase(measurements.universal.data.MeasurementsAnnualPeriodicUnionCache):

    def __init__(self, *measurement_list):
        super().__init__(*measurement_list)
        self.standard_deviation_concentration_noise_ratio = measurements.dop.pw.constants.DEVIATION_CONCENTRATION_NOISE_RATIO
        self.fill_strategy = 'point_average'


    ## standard_deviation_concentration_noise_ratio

    @property
    @util.cache.memory.method_decorator(dependency=('self.fill_strategy', 'self.min_measurements_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def concentration_standard_deviations_for_sample_lsm(self):
        return self.standard_deviation_concentration_noise_ratio * self.average_noise_standard_deviations_for_sample_lsm

    @property
    @util.cache.memory.method_decorator(dependency=('self.fill_strategy', 'self.min_measurements_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def concentration_standard_deviations(self):
        return self.standard_deviation_concentration_noise_ratio * self.average_noise_standard_deviations


    def _fill_strategy_for_kind(self, kind):
        if kind == 'concentration_standard_deviations':
            fill_strategy = self._fill_strategy_for_kind('average_noise_standard_deviations') + '_ratio_{:g}'.format(self.standard_deviation_concentration_noise_ratio)
        else:
            fill_strategy = super()._fill_strategy_for_kind(kind)
        return fill_strategy


class Measurements(MeasurementsBase):

    def __init__(self, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurement_list = [ measurement_class(min_measurements_correlation=min_measurements_correlation) for measurement_class in [MeasurementsLadolfi2002, MeasurementsLadolfi2004, MeasurementsYoshimura2007] ]
        super().__init__(*measurement_list)



class MeasurementsNearWaterLadolfi2002(measurements.universal.data.MeasurementsAnnualPeriodicNearWaterCache):

    def __init__(self, water_lsm=None, max_box_distance_to_water=0, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurements = MeasurementsLadolfi2002(min_measurements_correlation=min_measurements_correlation)
        super().__init__(measurements, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water)


class MeasurementsNearWaterLadolfi2004(measurements.universal.data.MeasurementsAnnualPeriodicNearWaterCache):

    def __init__(self, water_lsm=None, max_box_distance_to_water=0, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurements = MeasurementsLadolfi2004(min_measurements_correlation=min_measurements_correlation)
        super().__init__(measurements, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water)


class MeasurementsNearWaterYoshimura2007(measurements.universal.data.MeasurementsAnnualPeriodicNearWaterCache):

    def __init__(self, water_lsm=None, max_box_distance_to_water=0, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurements = MeasurementsYoshimura2007(min_measurements_correlation=min_measurements_correlation)
        super().__init__(measurements, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water)


class MeasurementsNearWater(MeasurementsBase):

    def __init__(self, water_lsm=None, max_box_distance_to_water=0, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurement_list = [ measurement_class(water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water, min_measurements_correlation=min_measurements_correlation) for measurement_class in [MeasurementsNearWaterLadolfi2002, MeasurementsNearWaterLadolfi2004, MeasurementsNearWaterYoshimura2007] ]
        super().__init__(*measurement_list)




import numpy as np

import logging
logger = logging.getLogger(__name__)

import measurements.land_sea_mask.lsm
import measurements.dop.box.regrid


def npy_or_save(npy_file):
    try:
        logger.debug('Loading data from {}.'.format(npy_file))
        data = np.load(npy_file)
    except (OSError, IOError):
        logger.debug('File {} does not exists. Calculating DOP data.'.format(npy_file))
        land_sea_mask = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=12)
        measurements.dop.box.regrid.save(land_sea_mask)
        data = np.load(npy_file)

    return data



def nobs():
    from measurements.dop.box.constants import NOBS_FILE
    data = npy_or_save(NOBS_FILE)
    return data


def variances():
    from measurements.dop.box.constants import VARIS_FILE
    data = npy_or_save(VARIS_FILE)
    return data


def means():
    from measurements.dop.box.constants import MEANS_FILE
    data = npy_or_save(MEANS_FILE)
    return data


import numpy as np

import util.io.netcdf
import measurements.po4.woa.data13.save

import util.logging
logger = util.logging.logger


## (raw) data from woa

def woa_data(data_name, divide_annual_values):
    from .constants import ANNUAL_FILE, MONTHLY_FILES

    def data_from_file(file, data_name):
        return util.io.netcdf.load(file, data_name).swapaxes(1,3)
        assert data.ndim == 4
        assert data.shape[:3] == (1, 360, 180)

    ## load annual data
    t_dim = len(MONTHLY_FILES)
    data = data_from_file(ANNUAL_FILE, data_name)
    # data = np.tile(data.T, t_dim).T
    data = np.repeat(data, t_dim, axis=0)
    if divide_annual_values:
        data = data / t_dim

    ## load monthly data
    for t in range(t_dim):
        data_monthly = data_from_file(MONTHLY_FILES[t], data_name)
        z_dim = data_monthly.shape[-1]
        data[t, :, :, :z_dim] = data_monthly

    ## return
    return data


def depth_values():
    from .constants import ANNUAL_FILE, DEPTH_DATANAME
    return util.io.netcdf.load(ANNUAL_FILE, DEPTH_DATANAME)


def measurement_data():
    from .constants import MEAN_DATANAME, NOBS_DATANAME, STANDARD_DEVIATION_DATANAME

    ## load data
    means = woa_data(MEAN_DATANAME, False).astype(np.float64)
    nobs = woa_data(NOBS_DATANAME, True)
    deviations = woa_data(STANDARD_DEVIATION_DATANAME, False).astype(np.float64)

    ## calculate sum of values and squares
    sum_of_values = means * nobs
    sum_of_squares = deviations**2 * (nobs - 1) + means**2

    mask = nobs == 1 & ~ means.mask
    sum_of_squares[mask] = means[mask]**2

    ## get measurement points
    mask = np.where(nobs > 0)
    points = np.array([mask_i.data for mask_i in mask]).swapaxes(0, 1)

    ## center months
    points[:, 0] += 0.5

    ## get corresponding depths
    depth = depth_values()
    points[:, 3] = depth[points[:, 3]]

    ### concatenate data
    measurement_data = np.concatenate([points, sum_of_values[mask].data[:,np.newaxis], nobs[mask].data[:,np.newaxis], sum_of_squares[mask].data[:,np.newaxis]], axis=1)

    return measurement_data



## regrided data

def npy_or_save(npy_file):
    try:
        logger.debug('Loading data from {}.'.format(npy_file))
        data = np.load(npy_file)
    except (OSError, IOError):
        logger.debug('File {} does not exists. Calculating PO4 woa data.'.format(npy_file))
        measurements.po4.woa.data13.save.measurements_to_metos_boxes()
        data = np.load(npy_file)

    return data


def nobs():
    from .constants import NOBS_FILE
    data = npy_or_save(NOBS_FILE)
    return data


def variances():
    from .constants import VARIANCES_FILE
    data = npy_or_save(VARIANCES_FILE)
    return data


def means():
    from .constants import MEANS_FILE
    data = npy_or_save(MEANS_FILE)
    return data



import numpy as np

DEFAULT_OPTIONS = {'start':50, 'increment_start':20, 'increment_value':10, 'increment_value_max':50, 'increment_step':1}

def diff(i, start=50, increment_start=20, increment_value=10, increment_value_max=50, increment_step=1):
    options = {'start':start, 'increment_start':increment_start, 'increment_value':increment_value, 'increment_value_max':increment_value_max, 'increment_step':increment_step}

    ## skip diff with no increase
    modulo = i % increment_step
    if modulo != 0:
        i = i - modulo

    ## calculate diff
    if i <= 0:
        return start
    else:
        previous_diff = diff(i - increment_step, **options)
        increment = min(increment_start  + (i / increment_step - 1) * increment_value, increment_value_max)
        return previous_diff + increment


def value(i, **options):
    if i <= 0:
        return 0
    else:
        return value(i-1, **options) + diff(i-1, **options)


def values_fix_length(length, **options):
    return list(map(lambda i: value(i, **options), range(length)))


def values_until_max(max_value, **options):
    values = []
    i = 0
    max_reached = False
    while not max_reached:
        value_i = value(i, **options)
        values.append(value_i)
        max_reached = max_value <= value_i
        i += 1

    return values


def values_TMM(max_value=9300, increment_step=1):
    start = 50 / increment_step
    increment_start = 20 / increment_step
    increment_value = 10 / increment_step
    increment_value_max = 50 / increment_step

    return values_until_max(max_value, start=start, increment_start=increment_start, increment_value=increment_value, increment_value_max=increment_value_max, increment_step=increment_step)




import os.path

import measurements.universal.dict
import measurements.po4.wod.constants


class MeasurementsDict(measurements.universal.dict.MeasurementsDict):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)

    def add_cruises(self, cruises):
        for cruise in cruises:
            t = cruise.time
            x = cruise.lon
            y = cruise.lat
            z = cruise.depth
            results = cruise.values.astype(float)
            assert len(z) == len(results)

            for i in range(results.size):
                index = (t, x, y, z[i])
                self.append_value(index, results[i])


    def save(self, file=measurements.po4.wod.constants.MEASUREMENTS_DICT_UNSORTED_FILE):
        super().save(file)

    @classmethod
    def load(cls, file=measurements.po4.wod.constants.MEASUREMENTS_DICT_UNSORTED_FILE):
        return super().load(file)



class MeasurementsUnsortedDict(MeasurementsDict):

    def __init__(self):
        super().__init__(sorted=False)



class MeasurementsSortedDict(MeasurementsDict):

    def __init__(self):
        super().__init__(sorted=False)

    def save(self, file=measurements.po4.wod.constants.MEASUREMENTS_DICT_SORTED_FILE):
        super().save(file)

    @classmethod
    def load(cls, file=measurements.po4.wod.constants.MEASUREMENTS_DICT_SORTED_FILE):
        return super().load(file)




import numpy as np
import os.path

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d     # for projection='3d'

from measurements.po4.wod.correlogram import estimation, model

import measurements.land_sea_mask.lsm
import measurements.po4.wod.mean.values
import measurements.po4.wod.mean.constants
import measurements.po4.wod.deviation.values
import measurements.po4.wod.deviation.constants

import util.plot



def data(calculation_kind, data_kind, lsm_kind, t_dim, dir='/tmp'):
    assert calculation_kind in ('sample', 'interpolated')
    assert data_kind in ('mean', 'deviation', 'total_deviation', 'average_noise_deviation', 'concentration_deviation')
    assert lsm_kind in ('TMM', 'WOA13', 'WOA13R')
    default_t_dim = 12
    assert t_dim in (1, default_t_dim)


    ## chose lsm
    if lsm_kind == 'TMM':
        sample_lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=default_t_dim)
    elif lsm_kind == 'WOA13':
        sample_lsm = measurements.land_sea_mask.lsm.LandSeaMaskWOA13(t_dim=default_t_dim)
    elif lsm_kind == 'WOA13R':
        sample_lsm = measurements.land_sea_mask.lsm.LandSeaMaskWOA13R(t_dim=default_t_dim)

    ## chose interpolator and min and max values
    if 'mean' in data_kind:
        values = measurements.po4.wod.mean.values
        constants = measurements.po4.wod.mean.constants
        min_values = constants.MIN_MEASUREMENTS

        v_min = 0
        v_max = 2.5

        histogram_step_size = 0.1
        histogram_v_max = 10
        depth_v_max = 2.5
        
        interpolator = values.InterpolatorConcentration(sample_lsm=sample_lsm, min_values=min_values)

    elif 'deviation' in data_kind:
        values = measurements.po4.wod.deviation.values
        constants = measurements.po4.wod.deviation.constants
        min_values = constants.MIN_MEASUREMENTS

        v_min = 0
        # if t_dim == 1:
        #     v_max = 0.3
        # elif t_dim == 4:
        #     v_max = 0.35
        # elif t_dim == 12:
        #     v_max = 0.4
        # elif t_dim == 48:
        #     v_max = 0.5
        v_max = 0.5

        histogram_step_size = 0.05
        histogram_v_max = 3.5
        depth_v_max = 0.3
        
        if data_kind == 'average_noise_deviation':
            interpolator = values.InterpolatorAverageNoise(sample_lsm=sample_lsm, min_values=min_values)
        elif data_kind == 'concentration_deviation':
            interpolator = values.InterpolatorConcentration(sample_lsm=sample_lsm, min_values=min_values)
        else:
            interpolator = values.InterpolatorTotal(sample_lsm=sample_lsm, min_values=min_values)

    ## chose date
    if calculation_kind == 'sample':
        data = interpolator.sample_data_for_lsm(sample_lsm)
    elif calculation_kind == 'interpolated':
        data = interpolator.interpolated_data_for_lsm(sample_lsm)
        volume_map = sample_lsm.volume_map
        data_depth_mean = np.nansum(data * volume_map, axis=(0,1,2)) / np.nansum(volume_map, axis=(0,1,2))
    
    contours = calculation_kind == 'interpolated'
    
    if t_dim == 1:
        data = data.mean(axis=0)

    ## prepare file names
    file_prefix = os.path.join(dir, 'po4_-_wod13_-_{data_kind}_-_{calculation_kind}_-_{sample_lsm}_-_min_values_{min_values}_-'.format(data_kind=data_kind, calculation_kind=calculation_kind, sample_lsm=sample_lsm, min_values=min_values))
    file_data = file_prefix + '.png'
    file_histogram = file_prefix + '_{}_histogram.png'.format(t_dim)
    file_depth_mean = file_prefix + '_depth_mean.png'.format()

    ## plot
    util.plot.data(data, file_data, no_data_value=np.inf, v_min=v_min, v_max=v_max, contours=contours, colorbar=not contours)
    util.plot.histogram(data[~np.isnan(data)], file_histogram, step_size=histogram_step_size, x_min=v_min, x_max=histogram_v_max, use_log_scale=True)
    if calculation_kind == 'interpolated':
        util.plot.line(sample_lsm.z_center, data_depth_mean, file_depth_mean, y_min=v_min, y_max=depth_v_max, line_color='b', line_width=3, xticks=np.arange(5)*2000)



def stationary_correlation(path='/tmp'):
    from measurements.constants import BASE_DIR
    c = np.load(os.path.join(BASE_DIR, 'po4/wod13/analysis/correlation/05_min_measurements_categorized_unsorted_correlations_stationary.npy'))

    dims = 4
    base_mask = np.all(np.isfinite(c), axis=1)

    ## plot correlation for changes in one and in two dims
    indices_list = [[(i,) for i in range(dim)], [(i,j) for i in range(dim) for j in range(dim) if i<j]]
    for indices in indices_list:
        for i in indices:
            ## calculate mask where only index i is not zero
            mask = base_mask.copy()
            for j in range(dims):
                if j not in i:
                    mask = np.logical_and(mask, c[:,j] == 0)

            ## scatter plot
            file = os.path.join(path, 'correlation_for_index_{}.png'.format(i))
            util.plot.scatter(c[mask][:,i], c[mask][:,-1], file, point_size=c[mask][:,-2])




def plot_correlogram(path='/tmp', show_model=True, min_measurements=1):

    direction_indices = estimation.get_direction_indices()

    ## get model
    if show_model:
        correlation_model = model.Correlation_Model()
        model_f = lambda d: correlation_model.correlation_by_distance(d)
    else:
        model_f = None

    for direction_index in direction_indices:
        ## get estimated data
        direction = estimation.get_direction(direction_index)
        shift = estimation.get_shift(direction_index, min_measurements=min_measurements)
        correlation = estimation.get_correlation(direction_index, min_measurements=min_measurements)
        number = estimation.get_number(direction_index, min_measurements=min_measurements)
        x_max = shift[len(shift)-1]

        ## plot setup
        fig = plt.figure(figsize=(15,10), dpi=150)
        plt.xlim((0, x_max))
        plt.ylim((-1, 1))

        title = ''
        if direction[0] > 0:
            days = int(int(np.round(1/direction[0])) / 365)
            if days == 1:
                title += '{} day '.format(days)
            else:
                title += '{} days '.format(days)
        if direction[1] > 0:
            if title != '':
                title += ', '
            title += '{}° longitude '.format(direction[1])
        if direction[2] > 0:
            if title != '':
                title += ', '
            title += '{}° latitude '.format(direction[2])
        if direction[3] > 0:
            if title != '':
                title += ', '
            title += '{}m depth'.format(direction[3])

        title = 'distance: ' + title
        plt.title(title)

        ## plot model
        if model_f is not None:
            x = np.arange(x_max)
            y = np.empty(x_max, np.float)
            for i in range(x_max):
                y[i] = model_f(direction * x[i])

            plt.plot(x, y, color='blue', linewidth=3)

        ## plot zero line
        plt.plot(np.arange(x_max+1), np.arange(x_max+1)*0, color='black', linewidth=2)

        ## plot estimated data
        size = np.log10(number ** 2) * 10
        plt.scatter(shift, correlation, size, color='red')

        ## set spine lines size
        util.plot.set_spine_line_size(fig, line_width=2)

        ## save plot
        file = os.path.join(path, 'po4_wod13_correlogram_min_measurements_' + str(min_measurements) + '_direction_' + str(direction_index) + '.png')
        plt.savefig(file, transparent=True)
        plt.close(fig)

        util.plot.trim(file)





import argparse

import measurements.po4.wod.plot.interface

import util.logging


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Plot sample or interpolated mean or deviation.')

    t_dims = (1, 4, 12, 48)
    lsms = ('TMM', 'WOA13R', 'WOA13')
    parser.add_argument('data', choices=('mean', 'total_deviation', 'concentration_deviation', 'average_noise_deviation'))
    parser.add_argument('calculation', choices=('sample', 'interpolated'))
    parser.add_argument('--lsm', choices=('all', )+lsms, default='all')
    parser.add_argument('--t_dim', choices=(0,)+t_dims, type=int, default=0)

    args = parser.parse_args()

    if args.t_dim != 0:
        t_dims = (args.t_dim,)
    if args.lsm != 'all':
        lsms = (args.lsm,)

    with util.logging.Logger():
        for lsm in lsms:
            for t_dim in t_dims:
                measurements.po4.wod.plot.interface.data(args.calculation, args.data, lsm, t_dim)



import argparse

import measurements.dop.pw.data
import measurements.po4.wod.data.values
import measurements.land_sea_mask.lsm
import measurements.util.plot

import util.logging


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Plotting interpolation data.')

    types = ('time', 'year', 'depth', 'space', 'space_tmm', 'space_woa13r', 'space_woa13')
    parser.add_argument('data', choices=('dop', 'po4'), help='The data to plot.')
    parser.add_argument('--type', choices=('all',)+types, default='all', help='The type of the plot.')

    args = parser.parse_args()

    data = args.data

    if data == 'dop':
        load_data = measurements.dop.pw.data.measurement_dict
        tick_power = None
        use_log_scale = False
    else:
        load_data = measurements.po4.wod.data.values.measurement_dict
        data = 'po4_-_wod13'
        tick_power = 3
        use_log_scale = True

    t_min = measurements.po4.wod.data.values.points()[:,0].min()
    t_max = measurements.po4.wod.data.values.points()[:,0].max()
    z_max = measurements.po4.wod.data.values.points()[:,3].max()

    type = args.type
    if type != 'all':
        types = (type,)


    with util.logging.Logger():
        for type in types:

            if type == 'time':
                measurements.util.plot.distribution_time(load_data(), file='/tmp/{}_-_distribution_time_-.png'.format(data), t_min=t_min, t_max=t_max, tick_power=tick_power)
            if type == 'year':
                for time_len in (365, 2880, 48):
                    measurements.util.plot.distribution_year(load_data(), file='/tmp/{}_-_distribution_year_{}_-.png'.format(data, time_len), time_step=1./time_len, tick_power=tick_power)
            if type == 'depth':
                measurements.util.plot.distribution_depth(load_data(), file='/tmp/{}_-_distribution_depth.png_-'.format(data), z_max=z_max, use_log_scale=use_log_scale)
            if type[:5] == 'space':
                type = type[6:]
                lsms = []
                if type == 'tmm' or type == '':
                    lsms.append(measurements.land_sea_mask.lsm.LandSeaMaskTMM())
                if type == 'woa13r' or type == '':
                    lsms.append(measurements.land_sea_mask.lsm.LandSeaMaskWOA13R())
                if type == 'woa13' or type == '':
                    lsms.append(measurements.land_sea_mask.lsm.LandSeaMaskWOA13())

                for lsm in lsms:
                    for t_dim in (12, 4, 1):
                        lsm.t_dim = t_dim
                        measurements.util.plot.distribution_space(load_data(), lsm=lsm, file='/tmp/{}_-_distribution_space_-.png'.format(data), use_log_scale=use_log_scale)




# measurements: a collection of functions to handel measurements
# Copyright (C) 2011-2016  Joscha Reimer jor@informatik.uni-kiel.de
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A setuptools based setup module.
https://packaging.python.org/en/latest/distributing.html
"""

import setuptools
import os.path

# Get the long description from the README file
readme_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.rst')
with open(readme_file, mode='r', encoding='utf-8') as f:
    long_description = f.read()

# Version string
def version():
    import setuptools_scm
    def empty_local_scheme(version):
        return ""
    return {'local_scheme': empty_local_scheme}


setuptools.setup(
    # Name
    name = 'measurements',
    
    # Desctiption
    description = 'measurements functions',
    long_description = long_description,
    
    # Keywords
    keywords = 'measurements functions',

    # Homepage
    url = 'https://github.com/jor-/measurements',

    # Author
    author = 'Joscha Reimer',
    author_email = 'jor@informatik.uni-kiel.de',

    # Version
    use_scm_version = version,

    # License
    license = 'GPLv3+',

    # Classifiers
    classifiers = [
        # Development Status
        'Development Status :: 3 - Alpha',

        # Intended Audience, Topic
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # Licence (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Supported Python versions
        'Programming Language :: Python :: 3',
    ],

    # Packages to install
    packages = setuptools.find_packages(),

    # Dependencies
    setup_requires = [
        'setuptools>=0.8',
        'pip>=1.4',
        'setuptools_scm',
    ],
    install_requires = [
        'numpy',
        'scipy',
        #'overrides',
        'utillib[cache,options,interpolate,cholmod]',
    ],
    extras_require = {
        'plot' : ['matplotlib', 'utillib[plot]'],
        'sorted_measurements_dict': ['utillib[sorted_multi_dict]'],
    },
)



import numpy as np

import measurements.po4.woa.data13.load
import measurements.land_sea_mask.lsm
# import measurements.util.map
import util.plot



def plot_sample_mean(file='/tmp/woa_po4_sample_mean.png', v_max=None, layer=None):
    data = measurements.po4.woa.data13.load.means()
    assert data.ndim == 4
    lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=len(data))
    data = lsm.apply_mask(data, land_value=np.inf)
#     for t_index in range(t_dim):
#         data[t_index] = measurements.util.map.apply_mask(lsm, data[t_index], land_value=np.inf)

    if layer is not None:
        data = data[:, :, :, layer]
        data = data.reshape(data.shape + (1,))
    util.plot.data(data, file, land_value=np.inf, no_data_value=np.nan, v_min=0, v_max=v_max)


def plot_sample_deviation(file='/tmp/woa_po4_sample_deviation.png', v_max=None, layer=None):
    data = measurements.po4.woa.data13.load.variances()**(1/2)
    assert data.ndim == 4
    lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=len(data))
    data = lsm.apply_mask(data, land_value=np.inf)
#     for t_index in range(len(data)):
#         data[t_index] = measurements.util.map.apply_mask(lsm, data[t_index], land_value=np.inf)

    if layer is not None:
        data = data[:, :, :, layer]
        data = data.reshape(data.shape + (1,))
    util.plot.data(data, file, land_value=np.inf, no_data_value=np.nan, v_min=0, v_max=v_max)


def plot_sample_nob(file='/tmp/woa_po4_sample_nob.png', v_max=None, layer=None):
    data = measurements.po4.woa.data13.load.nobs()
    assert data.ndim == 4
    lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=len(data))
    data = lsm.apply_mask(data, land_value=np.inf)
    if layer is not None:
        data = data[:, :, :, layer]
        data = data.reshape(data.shape + (1,))
    util.plot.data(data, file, no_data_value=0, v_min=1, v_max=v_max, use_log_norm=True)



import measurements.dop.pw.data
import measurements.po4.wod.data
import measurements.universal.data
import measurements.universal.constants
import measurements.land_sea_mask.lsm


def all_measurements(max_box_distance_to_water=None, min_measurements_correlation=None, tracers=None):
    if min_measurements_correlation is None:
        min_measurements_correlation = measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS
    
    measurements_collection = []
    
    if max_box_distance_to_water is None or max_box_distance_to_water == float('inf'):
        if tracers is None or 'po4' in tracers:
            measurements_collection.append(measurements.po4.wod.data.Measurements(min_measurements_correlation=min_measurements_correlation))
        if tracers is None or 'dop' in tracers:
            measurements_collection.append(measurements.dop.pw.data.Measurements(min_measurements_correlation=min_measurements_correlation))
    else:
        water_lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM()
        if tracers is None or 'po4' in tracers:
            measurements_collection.append(measurements.po4.wod.data.MeasurementsNearWater(min_measurements_correlation=min_measurements_correlation, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water))
        if tracers is None or 'dop' in tracers:
            measurements_collection.append(measurements.dop.pw.data.MeasurementsNearWater(min_measurements_correlation=min_measurements_correlation, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water))
    
    measurements_collection = measurements.universal.data.MeasurementsCollectionCache(*measurements_collection)
    return measurements_collection



import bisect
import logging
import numpy as np

import measurements.po4.woa.data13.load
import measurements.util.regrid

import util.math.interpolate



def measurements_to_land_sea_mask(land_sea_mask, z_values):
    from .constants import T_RANGE, X_RANGE, Y_RANGE, ANNUAL_THRESHOLD
    from .constants import VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR, VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR, VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND
    from measurements.po4.constants import DEVIATION_MIN_MEASUREMENTS, DEVIATION_MIN_VALUE

    ## load measurement data and regrid them
    measurement_data = measurements.po4.woa.data13.load.measurement_data()
    (means, nobs, variances) = measurements.util.regrid.measurements_to_land_sea_mask(measurement_data, land_sea_mask, t_range=T_RANGE, x_range=X_RANGE, y_range=Y_RANGE)
    assert means.ndim == 4 and nobs.ndim == 4 and variances.ndim == 4
    assert means.shape == nobs.shape == variances.ndim

    ## remove variances with to few measurements
    ANNUAL_THRESHOLD_INDEX = bisect.bisect_right(z_values, ANNUAL_THRESHOLD)
    variances[:,:,:,:ANNUAL_THRESHOLD_INDEX][nobs[:,:,:,:ANNUAL_THRESHOLD_INDEX] < DEVIATION_MIN_MEASUREMENTS] = np.inf
    variances[:,:,:,ANNUAL_THRESHOLD_INDEX:][nobs[:,:,:,ANNUAL_THRESHOLD_INDEX:] < DEVIATION_MIN_MEASUREMENTS / land_sea_mask.t_dim] = np.inf

    ## set variances below lower bound to lower bound
    variances[variances < DEVIATION_MIN_VALUE**2] = DEVIATION_MIN_VALUE**2


    ## interpolate missing variances
    data_index_t, data_index_x, data_index_y, data_index_z = np.where(nobs >= DEVIATION_MIN_MEASUREMENTS)
    data_points_indices = (data_index_t, data_index_x, data_index_y, data_index_z)
    data_points = np.array(data_points_indices).swapaxes(0,1)
    data_values = variances[data_points_indices]

    interpolation_index_t, interpolation_index_x, interpolation_index_y, interpolation_index_z = np.where(nobs < DEVIATION_MIN_MEASUREMENTS)
    interpolation_points_indices = (interpolation_index_t, interpolation_index_x, interpolation_index_y, interpolation_index_z)
    interpolation_points = np.array(interpolation_points_indices).swapaxes(0,1)

    logging.debug('Interpolating variance for {} points.'.format(len(interpolation_points)))

    interpolator = util.math.interpolate.Periodic_Interpolator(data_points, data_values, point_range_size=variances.shape, wrap_around_amount=VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND, number_of_linear_interpolators=VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR, single_overlapping_amount_linear_interpolators=VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR/VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR)

    variances[interpolation_points_indices] = interpolator.interpolate(interpolation_points)

    assert np.all(np.isfinite(np.logical_not(np.isnan(variances))))


    ## return regrided values
    return (means, nobs, variances)




import numpy as np

import logging
logger = logging.getLogger(__name__)

import measurements.po4.woa.data09.regrid


def npy_or_save(npy_file):
    try:
        logger.debug('Loading data from {}.'.format(npy_file))
        data = np.load(npy_file)
    except (OSError, IOError):
        logger.debug('File {} does not exists. Calculating PO4 data.'.format(npy_file))
        measurements.po4.woa.data09.regrid.save()
        data = np.load(npy_file)

    return data


def nobs():
    from .constants import NOBS_FILE
    data = npy_or_save(NOBS_FILE)
    return data



def variances():
    from .constants import VARIS_FILE
    data = npy_or_save(VARIS_FILE)
    return data



def means():
    from .constants import MEANS_FILE
    data = npy_or_save(MEANS_FILE)
    return data


import os.path

from measurements.constants import BASE_DIR


## PO4
WOA_BASE_DIR = os.path.join(BASE_DIR, 'po4/woa09')
NOBS_NETCDF_MONTHLY_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_monthly_nobs.cdf')
NOBS_NETCDF_ANNUAL_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_annual_nobs.cdf')
NOBS_NETCDF_DATANAME = 'PO4NOBS'

VARIS_NETCDF_MONTHLY_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_monthly_vari.cdf')
VARIS_NETCDF_ANNUAL_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_annual_vari.cdf')
VARIS_NETCDF_DATANAME = 'PO4VARI'

MEANS_NETCDF_MONTHLY_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_monthly_mean.cdf')
MEANS_NETCDF_ANNUAL_FILE = os.path.join(WOA_BASE_DIR, 'woa09_po4_2.8x2.8_annual_mean.cdf')
MEANS_NETCDF_DATANAME = 'PO4MEAN'

NOBS_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_nobs.npy')
VARIS_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_vari.npy')
MEANS_FILE = os.path.join(WOA_BASE_DIR, 'po4_2.8x2.8_monthly_mean.npy')

ANNUAL_THRESHOLD = 500

VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR = 1
VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR = 0
VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND = (0.1, 0.1, 0, 0)


#
# ## DOP
# YOSHIMURA_DOP_MEASUREMENT_FILE = os.path.join(BASE_DIR, 'dop/Yoshimura2007/Yoshimura2007_prepared.txt')
# LADOLFI_2002_DOP_MEASUREMENT_FILE = os.path.join(BASE_DIR, 'dop/Ladolfi2002/CD139_DOP_prepared.txt')
# LADOLFI_2004_DOP_MEASUREMENT_FILE = os.path.join(BASE_DIR, 'dop/Ladolfi2004/D279_DOP_prepared.txt')
#
# DOP_NOBS = os.path.join(BASE_DIR, 'dop/dop_2.8x2.8_monthly_nobs.npy')
# DOP_VARIS = os.path.join(BASE_DIR, 'dop/dop_2.8x2.8_monthly_vari.npy')
# DOP_MEANS = os.path.join(BASE_DIR, 'dop/dop_2.8x2.8_monthly_mean.npy')
#
#
# ## BOTH
# NOBS = os.path.join(BASE_DIR, 'dop_po4_2.8x2.8_monthly_nobs.npy')
# VARIS = os.path.join(BASE_DIR, 'dop_po4_2.8x2.8_monthly_vari.npy')
# MEANS = os.path.join(BASE_DIR, 'dop_po4_2.8x2.8_monthly_mean.npy')




import os.path

import measurements.all.box.data as data
from measurements.all.box.constants import WOA_BASE_DIR

import measurements.all.box.constants as all
import measurements.dop.box.constants as dop
import measurements.po4.woa.data13.constants as po4

import util.logging
import util.io.fs

log_file = os.path.join(WOA_BASE_DIR, 'save_values_' + str(po4.VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND) + '_' + str(po4.VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR) + '_' + str(po4.VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR) + '.log')

with util.logging.Logger(log_file=log_file, disp_stdout=not __name__ == "__main__"):
    for file in (all.NOBS_FILE, all.VARIS_FILE, all.MEANS_FILE, po4.NOBS_FILE, po4.VARIS_FILE, po4.MEANS_FILE, dop.NOBS_FILE, dop.VARIS_FILE, dop.MEANS_FILE):
        util.io.fs.remove_file(file, not_exist_okay=True)

    nobs = data.nobs()
    means = data.means()
    varis = data.variances()

